#include "AesFile.h"
#include "mbedtls/aes.h"



#pragma pack(push, 1)
struct FileHeader32
{
	uint8_t magic16[AesFile::MAGIC_SIZE];	//[16]
	uint32_t alignedDataSize;				//[4]
	uint32_t realDataSize;					//[4]
	uint8_t reserved[8];					//[8]
};
#pragma pack(pop)



std::vector<uint8_t> AesFile::encryptFileData(const KeyFile &keyFile, const std::vector<uint8_t> &plainData)
{
	const size_t alignedSize = (plainData.size() % 16 == 0) ? plainData.size() : (plainData.size() + (16 - (plainData.size() % 16)));

	std::vector<uint8_t> alignedPlainData(sizeof(FileHeader32) + alignedSize, 0);
	FileHeader32 *pHeader = (FileHeader32*)alignedPlainData.data();
	for (size_t i = 0; i < plainData.size(); ++i)
		alignedPlainData[sizeof(FileHeader32) + i] = plainData[i];

	memcpy(pHeader->magic16, keyFile.magic, sizeof(pHeader->magic16));
	pHeader->alignedDataSize = (uint32_t)alignedSize;
	pHeader->realDataSize = (uint32_t)plainData.size();
	memset(pHeader->reserved, 0, sizeof(pHeader->reserved));

	return encryptData(keyFile.key, keyFile.iv, alignedPlainData);
}

bool AesFile::decryptFileData(const KeyFile & keyFile, const std::vector<uint8_t>& cryptedData, std::vector<uint8_t>& plainData)
{
	plainData.clear();

	if ((cryptedData.size() >= sizeof(FileHeader32)) && (cryptedData.size() % 16 == 0))
	{
		std::vector<uint8_t> header_data;
		if (decryptData(keyFile.key, keyFile.iv, cryptedData, sizeof(FileHeader32), header_data) && header_data.size() == sizeof(FileHeader32))
		{
			const FileHeader32 *pHeader = (const FileHeader32*)header_data.data();
			if (!memcmp(pHeader->magic16, keyFile.magic, 16))
			{
				const size_t alignedDataSize = (size_t)pHeader->alignedDataSize;
				if (cryptedData.size() == (alignedDataSize + sizeof(FileHeader32)))
				{
					std::vector<uint8_t> alignedPlainData;
					if (decryptData(keyFile.key, keyFile.iv, cryptedData, cryptedData.size(), alignedPlainData) && alignedPlainData.size() == cryptedData.size())
					{
						const size_t realDataSize = (size_t)pHeader->realDataSize;
						if (realDataSize <= alignedDataSize)
						{
							plainData.resize(realDataSize, 0);
							for (size_t i = 0; i < realDataSize; ++i)
								plainData[i] = alignedPlainData[sizeof(FileHeader32) + i];

							return true;
						}
					}
				}
			}
		}
	}

	return false;
}

AesFile::AesFile(const KeyFile & keyFile)
	: mKeyFile(keyFile)
{
//	mKeyFile.print();
}

AesFile::~AesFile()
{
}

std::vector<uint8_t> AesFile::encryptFileData(const std::vector<uint8_t>& plainData)
{
	return encryptFileData(mKeyFile, plainData);
}

bool AesFile::decryptFileData(const std::vector<uint8_t>& cryptedData, std::vector<uint8_t>& plainData)
{
	return decryptFileData(mKeyFile, cryptedData, plainData);
}

void AesFile::encryptAES256(const uint8_t key[32], const uint8_t iv[16], const uint8_t *data16, size_t data_size16, uint8_t *out16)
{
	mbedtls_aes_context ctx = {};
	mbedtls_aes_init(&ctx);
	mbedtls_aes_setkey_enc(&ctx, key, 256);

	uint8_t tmp_iv[16] = {};
	memcpy(tmp_iv, iv, 16);

	mbedtls_aes_crypt_cbc(&ctx, MBEDTLS_AES_ENCRYPT, data_size16, tmp_iv, data16, out16);

	mbedtls_aes_free(&ctx);
}

void AesFile::decryptAES256(const uint8_t key[32], const uint8_t iv[16], const uint8_t *data16, size_t data_size16, uint8_t *out16)
{
	mbedtls_aes_context ctx = {};
	mbedtls_aes_init(&ctx);
	mbedtls_aes_setkey_dec(&ctx, key, 256);

	uint8_t tmp_iv[16] = {};
	memcpy(tmp_iv, iv, 16);

	mbedtls_aes_crypt_cbc(&ctx, MBEDTLS_AES_DECRYPT, data_size16, tmp_iv, data16, out16);

	mbedtls_aes_free(&ctx);
}

std::vector<uint8_t> AesFile::encryptData(const uint8_t key[32], const uint8_t iv[16], const std::vector<uint8_t> &plainData)
{
	std::vector<uint8_t> alignedPlainData(plainData);

	if (alignedPlainData.size() < 16)
		alignedPlainData.resize(16, 0);

	while (alignedPlainData.size() % 16 != 0)
		alignedPlainData.push_back(0);

	std::vector<uint8_t> cryptedData(alignedPlainData.size(), 0);

	const uint8_t *inData16 = alignedPlainData.data();
	uint8_t *outData16 = cryptedData.data();
	size_t data_size16 = alignedPlainData.size();
	encryptAES256(key, iv, inData16, data_size16, outData16);

	return cryptedData;
}

bool AesFile::decryptData(const uint8_t key[32], const uint8_t iv[16], const std::vector<uint8_t> &cryptedData, size_t realDataSize, std::vector<uint8_t> &plainData)
{
	plainData.clear();

	if ((cryptedData.size() >= 16) && (cryptedData.size() % 16 == 0) && (realDataSize <= cryptedData.size()))
	{
		plainData.resize(cryptedData.size(), 0);

		const uint8_t *inData16 = cryptedData.data();
		size_t data_size16 = cryptedData.size();
		uint8_t *outData16 = plainData.data();
		decryptAES256(key, iv, inData16, data_size16, outData16);

		if (plainData.size() > realDataSize)
			plainData.resize(realDataSize);

		return true;
	}

	return false;
}
