#include "CDBField.h"
#include "Common.h"



enum RawFieldType {RFT_UNKNOWN=0, RFT_INTEGER=0xFF11, RFT_TEXT=0xFF22};


CDBFieldInteger::CDBFieldInteger()
	: mNull(true), mValue(0)
{
}

CDBFieldInteger::CDBFieldInteger(int64_t value)
	: mNull(false), mValue(value)
{
}

CDBFieldInteger::~CDBFieldInteger()
{
}

bool CDBFieldInteger::isNull() const
{
	return mNull;
}

void CDBFieldInteger::setNull()
{
	mNull = true;
	mValue = 0;
}

void CDBFieldInteger::setValue(int64_t value)
{
	mNull = false;
	mValue = value;
}

int64_t CDBFieldInteger::getValue(int64_t def) const
{
	if (!mNull)
		return mValue;
	return def;
}

#pragma pack(push, 1)
struct HeaderRawInteger
{
	uint32_t ftype;
	uint32_t fnull;
	uint64_t fval;
};
#pragma pack(pop)

std::vector<uint8_t> CDBFieldInteger::toRawData() const
{
	HeaderRawInteger header = {};
	header.ftype = static_cast<uint32_t>(RFT_INTEGER);
	header.fnull = mNull ? static_cast<uint32_t>(1) : static_cast<uint32_t>(0);
	header.fval = static_cast<uint64_t>(mValue);

	std::vector<uint8_t> raw(sizeof(header), 0);
	memcpy(raw.data(), &header, raw.size());

	return raw;
}

CDBFieldInteger CDBFieldInteger::fromRawData(const std::vector<uint8_t>& rawData)
{
	if (rawData.size() == sizeof(HeaderRawInteger))
	{
		const HeaderRawInteger *pHeader = (const HeaderRawInteger *)rawData.data();
		if (pHeader->ftype == static_cast<uint32_t>(RFT_INTEGER))
		{
			if (pHeader->fnull == static_cast<uint32_t>(0))
			{
				const int64_t val = static_cast<int64_t>(pHeader->fval);
				return CDBFieldInteger(val);
			}
		}
	}

	return CDBFieldInteger();
}



// TEXT



CDBFieldText::CDBFieldText()
	: mNull(true), mValue(wxString())
{
}

CDBFieldText::CDBFieldText(wxString value)
	: mNull(false), mValue(value)
{
}

CDBFieldText::~CDBFieldText()
{
}

bool CDBFieldText::isNull() const
{
	return mNull;
}

void CDBFieldText::setNull()
{
	mNull = true;
	mValue = wxString();
}

void CDBFieldText::setValue(wxString value)
{
	mNull = false;
	mValue = value;
}

wxString CDBFieldText::getValue(const wxString & def) const
{
	if (!mNull)
		return mValue;
	return def;
}

#pragma pack(push, 1)
struct HeaderRawText
{
	uint32_t ftype;
	uint32_t fnull;
	uint32_t fstr_len;
};
#pragma pack(pop)

std::vector<uint8_t> CDBFieldText::toRawData() const
{
	const std::vector<uint8_t> rawStr = wxStringToRawData(mValue);
	const size_t size_rawStr = rawStr.size();

	HeaderRawText header = {};
	header.ftype = static_cast<uint32_t>(RFT_TEXT);
	header.fnull = mNull ? static_cast<uint32_t>(1) : static_cast<uint32_t>(0);
	header.fstr_len = (uint32_t)size_rawStr;

	std::vector<uint8_t> raw(sizeof(header) + size_rawStr, 0);
	memcpy(raw.data(), &header, sizeof(header));
	memcpy(raw.data() + sizeof(header), rawStr.data(), size_rawStr);

	return raw;
}

CDBFieldText CDBFieldText::fromRawData(const std::vector<uint8_t>& rawData)
{
	if (rawData.size() >= sizeof(HeaderRawText))
	{
		const HeaderRawText *pHeader = (const HeaderRawText *)rawData.data();
		if (pHeader->ftype == static_cast<uint32_t>(RFT_TEXT))
		{
			if (pHeader->fnull == static_cast<uint32_t>(0))
			{
				const size_t str_size = (size_t)pHeader->fstr_len;
				if (str_size > 0)
				{
					if (rawData.size() == (sizeof(HeaderRawText) + str_size))
					{
						std::vector<uint8_t> rawStr(str_size, 0);
						memcpy(rawStr.data(), rawData.data() + sizeof(HeaderRawText), str_size);
						
						const wxString s = wxStringFromRawData(rawStr);
						return CDBFieldText(s);
					}
				}
				else
					return CDBFieldText(wxString());
			}
		}
	}

	return CDBFieldText();
}
