#include "CDBMemDataBase.h"


enum RawDBType { RDT_UNKNOWN = 0, RDT_SERVER_V1 = 0xCC01 };


int64_t CDBMemDataBase::tableLocalPathsAddLine(const CDBTablePathLine & pathLine)
{
	return tableLocalPaths.addLine(pathLine);
}

std::vector<CIDLinePath> CDBMemDataBase::tableLocalPathsReadAllLines()
{
	return tableLocalPaths.getAllLines();
}

int64_t CDBMemDataBase::tableRemotePathsAddLine(const CDBTablePathLine & pathLine)
{
	return tableRemotePaths.addLine(pathLine);
}

std::vector<CIDLinePath> CDBMemDataBase::tableRemotePathsReadAllLines()
{
	return tableRemotePaths.getAllLines();
}

int64_t CDBMemDataBase::tableServersAddLine(const CDBTableServerLine & serverLine)
{
	return tableServers.addLine(serverLine);
}

std::vector<CIDLineServer> CDBMemDataBase::tableServersReadAllLines()
{
	return tableServers.getAllLines();
}

int64_t CDBMemDataBase::tableFilesAddLine(const CDBTableFileLine & fileLine)
{
	const int64_t file_id = tableFiles.addLine(fileLine);
	if (file_id > 0)
	{
		const int64_t server_id = fileLine.server_id.getValue();

		CIDLineFile idLine;
		idLine.first = file_id;
		idLine.second = fileLine;

		if (mMapServerIdToFileLines.find(server_id) != mMapServerIdToFileLines.end())
			mMapServerIdToFileLines[server_id].push_back(idLine);
		else
		{
			mMapServerIdToFileLines[server_id] = std::vector<CIDLineFile>();
			mMapServerIdToFileLines[server_id].push_back(idLine);
		}
	}

	return file_id;
}

std::vector<CIDLineFile> CDBMemDataBase::tableFilesReadAllLinesByServerId(int64_t serverId)
{
	if (mMapServerIdToFileLines.find(serverId) != mMapServerIdToFileLines.end())
		return mMapServerIdToFileLines[serverId];

	return std::vector<CIDLineFile>();
}



#pragma pack(push, 1)
struct MemDBHeader
{
	uint32_t dbtype;
	uint32_t count_tables;
};
#pragma pack(pop)

std::vector<uint8_t> packRawTablesToRawData(const std::vector<std::vector<uint8_t> > &rawTables, RawDBType dbType)
{
	const size_t tables_count = rawTables.size();

	std::vector<uint8_t> rawData(sizeof(MemDBHeader), 0);

	MemDBHeader* pHeader = (MemDBHeader*)rawData.data();
	pHeader->dbtype = (uint32_t)dbType;
	pHeader->count_tables = (uint32_t)tables_count;

	if (tables_count > 0)
	{
		for (const std::vector<uint8_t> &table : rawTables)
		{
			const size_t table_size = table.size();

			std::vector<uint8_t> packed_table(sizeof(uint32_t), 0);
			uint32_t* pSizeTable = (uint32_t*)packed_table.data();
			*pSizeTable = (uint32_t)table_size;

			if (table_size > 0)
				packed_table.insert(packed_table.end(), table.begin(), table.end());

			rawData.insert(rawData.end(), packed_table.begin(), packed_table.end());
		}
	}

	return rawData;
}

std::vector<std::vector<uint8_t> > unpackRawTablesFromRawData(const std::vector<uint8_t> &rawData, RawDBType dbType)
{
	std::vector<std::vector<uint8_t> > rawTables;

	if (rawData.size() >= sizeof(MemDBHeader))
	{
		const MemDBHeader* pHeader = (const MemDBHeader*)rawData.data();
		const uint32_t db_type = (uint32_t)pHeader->dbtype;
		const size_t tables_count = (size_t)pHeader->count_tables;

		if (db_type == (uint32_t)dbType)
		{
			if (tables_count > 0)
			{
				size_t cur_off = sizeof(MemDBHeader);
				size_t rest_bytes = rawData.size() - sizeof(MemDBHeader);

				for (size_t i = 0; i < tables_count; ++i)
				{
					if (rest_bytes >= sizeof(uint32_t))
					{
						const uint32_t* pTableSize = (const uint32_t*)(rawData.data() + cur_off);
						const size_t table_size = (size_t)(*pTableSize);

						cur_off += sizeof(uint32_t);
						rest_bytes -= sizeof(uint32_t);

						if (table_size > 0)
						{
							if (rest_bytes >= table_size)
							{
								std::vector<uint8_t> rawTable(table_size, 0);
								memcpy(rawTable.data(), rawData.data() + cur_off, table_size);

								cur_off += table_size;
								rest_bytes -= table_size;

								rawTables.push_back(rawTable);
							}
							else
								break;
						}
						else
							rawTables.push_back(std::vector<uint8_t>());
					}
					else
						break;
				}
			}
		}
	}

	return rawTables;
}



std::vector<uint8_t> CDBMemDataBase::toRawData() const
{
	std::vector<std::vector<uint8_t> > rawTables;
	
	rawTables.push_back(tableLocalPaths.toRawData());
	rawTables.push_back(tableRemotePaths.toRawData());
	rawTables.push_back(tableServers.toRawData());
	rawTables.push_back(tableFiles.toRawData());

	const std::vector<uint8_t> rawData = packRawTablesToRawData(rawTables, RDT_SERVER_V1);

	return rawData;
}

CDBMemDataBase CDBMemDataBase::fromRawData(const std::vector<uint8_t>& rawData)
{
	std::vector<std::vector<uint8_t> > rawTables = unpackRawTablesFromRawData(rawData, RDT_SERVER_V1);

	if (rawTables.size() == 4)
	{
		CDBMemDataBase memDB;
		
		memDB.tableLocalPaths = CDBTablePath::fromRawData(rawTables[0]);
		memDB.tableRemotePaths = CDBTablePath::fromRawData(rawTables[1]);
		memDB.tableServers = CDBTableServer::fromRawData(rawTables[2]);
		memDB.tableFiles = CDBTableFile::fromRawData(rawTables[3]);

		memDB.updateCache();

		return memDB;
	}

	return CDBMemDataBase();
}

void CDBMemDataBase::updateCache()
{
	mMapServerIdToFileLines.clear();

	std::vector<CIDLineFile> file_lines = tableFiles.getAllLines();
	for (const CIDLineFile &file_line : file_lines)
	{
		const int64_t server_id = file_line.second.server_id.getValue();
		if (server_id > 0)
		{
			if (mMapServerIdToFileLines.find(server_id) != mMapServerIdToFileLines.end())
				mMapServerIdToFileLines[server_id].push_back(file_line);
			else
			{
				mMapServerIdToFileLines[server_id] = std::vector<CIDLineFile>();
				mMapServerIdToFileLines[server_id].push_back(file_line);
			}
		}
	}
}
