#include "CDBTables.h"



enum RawTableType { RTT_UNKNOWN = 0, RTT_SERVER = 0xDD11, RTT_FILE = 0xDD22, RTT_PATH = 0xDD33 };


std::vector<uint8_t> tableRawLinesToRawData(RawTableType tableType, int64_t idCounter, const std::vector<std::pair<int64_t, std::vector<uint8_t> > > &rawLines);
std::vector<std::pair<int64_t, std::vector<uint8_t> > > tableRawLinesFromRawData(RawTableType tableType, const std::vector<uint8_t> &rawData, int64_t &idCounter);



#pragma pack(push, 1)
struct TableHeader
{
	uint32_t type_table;
	int64_t id_counter;
	uint64_t count_rows;
};
#pragma pack(pop)


#pragma pack(push, 1)
struct RowHeader
{
	int64_t row_id;
	uint32_t row_size;
};
#pragma pack(pop)


std::vector<uint8_t> tableRawLinesToRawData(RawTableType tableType, int64_t idCounter, const std::vector<std::pair<int64_t, std::vector<uint8_t> > > &rawLines)
{
	const size_t lines_count = rawLines.size();

	std::vector<uint8_t> rawData(sizeof(TableHeader), 0);
	
	TableHeader* pHeader = (TableHeader*)rawData.data();
	pHeader->type_table = (uint32_t)tableType;
	pHeader->id_counter = (int64_t)idCounter;
	pHeader->count_rows = (uint64_t)lines_count;

	if (lines_count > 0)
	{
		for (const std::pair<int64_t, std::vector<uint8_t> > &row : rawLines)
		{
			const int64_t row_id = row.first;
			const std::vector<uint8_t> row_data = row.second;
			const size_t row_size = row_data.size();

			std::vector<uint8_t> packedRow(sizeof(RowHeader), 0);
			RowHeader* phRow = (RowHeader*)packedRow.data();
			phRow->row_id = (int64_t)row_id;
			phRow->row_size = (uint32_t)row_size;

			if (row_size > 0)
				packedRow.insert(packedRow.end(), row_data.begin(), row_data.end());

			rawData.insert(rawData.end(), packedRow.begin(), packedRow.end());
		}
	}

	return rawData;
}

std::vector<std::pair<int64_t, std::vector<uint8_t> > > tableRawLinesFromRawData(RawTableType tableType, const std::vector<uint8_t> &rawData, int64_t &idCounter)
{
	idCounter = 0;

	std::vector<std::pair<int64_t, std::vector<uint8_t> > > rawLines;

	if (rawData.size() >= sizeof(TableHeader))
	{
		const TableHeader* pHeader = (const TableHeader*)rawData.data();
		const uint32_t ttype = (uint32_t)pHeader->type_table;
		const int64_t idCnt = (int64_t)pHeader->id_counter;
		const size_t row_count = (size_t)pHeader->count_rows;

		if (ttype == (uint32_t)tableType)
		{
			idCounter = idCnt;

			if (row_count > 0)
			{
				size_t cur_off = sizeof(TableHeader);
				size_t rest_bytes = rawData.size() - sizeof(TableHeader);

				for (size_t i = 0; i < row_count; ++i)
				{
					if (rest_bytes >= sizeof(RowHeader))
					{
						const RowHeader* phRow = (const RowHeader*)(rawData.data() + cur_off);
						const int64_t row_id = (int64_t)phRow->row_id;
						const size_t row_size = (size_t)phRow->row_size;

						cur_off += sizeof(RowHeader);
						rest_bytes -= sizeof(RowHeader);

						if (row_size > 0)
						{
							if (rest_bytes >= row_size)
							{
								std::vector<uint8_t> rawRow(row_size, 0);
								memcpy(rawRow.data(), rawData.data() + cur_off, row_size);

								cur_off += row_size;
								rest_bytes -= row_size;

								rawLines.push_back(std::pair<int64_t, std::vector<uint8_t> >(row_id, rawRow));
							}
							else
								break;
						}
						else
							rawLines.push_back(std::pair<int64_t, std::vector<uint8_t> >(row_id, std::vector<uint8_t>()));
					}
					else
						break;
				}
			}
		}
	}

	return rawLines;
}




// CDBTableServer

std::vector<uint8_t> CDBTableServer::toRawData() const
{
	const std::vector<CIDLineServer> allLines = getAllLines();

	std::vector<std::pair<int64_t, std::vector<uint8_t> > > lines;

	for (const CIDLineServer &idLine : allLines)
	{
		const int64_t id = idLine.first;
		const CDBTableServerLine tline = idLine.second;

		const std::vector<uint8_t> rawLine = tline.toRawData();

		std::pair<int64_t, std::vector<uint8_t> > elem;
		elem.first = id;
		elem.second = rawLine;

		lines.push_back(elem);
	}

	const std::vector<uint8_t> rawData = tableRawLinesToRawData(RTT_SERVER, getIDCounter(), lines);

	return rawData;
}

CDBTableServer CDBTableServer::fromRawData(const std::vector<uint8_t>& rawData)
{
	int64_t idCnt = 0;
	std::vector<std::pair<int64_t, std::vector<uint8_t> > > rawLines = tableRawLinesFromRawData(RTT_SERVER, rawData, idCnt);

	CDBTableServer table;
	table.setIDCounter(idCnt);

	for (const std::pair<int64_t, std::vector<uint8_t> > &rawLine : rawLines)
	{
		const int64_t id = rawLine.first;
		const std::vector<uint8_t> rawTLine = rawLine.second;

		const CDBTableServerLine tline = CDBTableServerLine::fromRawData(rawTLine);

		table.mMapLines[id] = tline;
	}

	return table;
}



// CDBTableFile

std::vector<uint8_t> CDBTableFile::toRawData() const
{
	const std::vector<CIDLineFile> allLines = getAllLines();

	std::vector<std::pair<int64_t, std::vector<uint8_t> > > lines;

	for (const CIDLineFile &idLine : allLines)
	{
		const int64_t id = idLine.first;
		const CDBTableFileLine tline = idLine.second;

		const std::vector<uint8_t> rawLine = tline.toRawData();

		std::pair<int64_t, std::vector<uint8_t> > elem;
		elem.first = id;
		elem.second = rawLine;

		lines.push_back(elem);
	}

	const std::vector<uint8_t> rawData = tableRawLinesToRawData(RTT_FILE, getIDCounter(), lines);

	return rawData;
}

CDBTableFile CDBTableFile::fromRawData(const std::vector<uint8_t>& rawData)
{
	int64_t idCnt = 0;
	std::vector<std::pair<int64_t, std::vector<uint8_t> > > rawLines = tableRawLinesFromRawData(RTT_FILE, rawData, idCnt);

	CDBTableFile table;
	table.setIDCounter(idCnt);

	for (const std::pair<int64_t, std::vector<uint8_t> > &rawLine : rawLines)
	{
		const int64_t id = rawLine.first;
		const std::vector<uint8_t> rawTLine = rawLine.second;

		const CDBTableFileLine tline = CDBTableFileLine::fromRawData(rawTLine);

		table.mMapLines[id] = tline;
	}

	return table;
}




// CDBTablePath

std::vector<uint8_t> CDBTablePath::toRawData() const
{
	const std::vector<CIDLinePath> allLines = getAllLines();

	std::vector<std::pair<int64_t, std::vector<uint8_t> > > lines;

	for (const CIDLinePath &idLine : allLines)
	{
		const int64_t id = idLine.first;
		const CDBTablePathLine tline = idLine.second;

		const std::vector<uint8_t> rawLine = tline.toRawData();

		std::pair<int64_t, std::vector<uint8_t> > elem;
		elem.first = id;
		elem.second = rawLine;

		lines.push_back(elem);
	}

	const std::vector<uint8_t> rawData = tableRawLinesToRawData(RTT_PATH, getIDCounter(), lines);

	return rawData;
}

CDBTablePath CDBTablePath::fromRawData(const std::vector<uint8_t>& rawData)
{
	int64_t idCnt = 0;
	std::vector<std::pair<int64_t, std::vector<uint8_t> > > rawLines = tableRawLinesFromRawData(RTT_PATH, rawData, idCnt);

	CDBTablePath table;
	table.setIDCounter(idCnt);

	for (const std::pair<int64_t, std::vector<uint8_t> > &rawLine : rawLines)
	{
		const int64_t id = rawLine.first;
		const std::vector<uint8_t> rawTLine = rawLine.second;

		const CDBTablePathLine tline = CDBTablePathLine::fromRawData(rawTLine);

		table.mMapLines[id] = tline;
	}

	return table;
}
