#include "CryptFiles.h"
#include "msgbox.h"
#include <wx/buffer.h>
#include <stdio.h>
#include <fstream>
#include <utility>
#include <wx/file.h>
#include <wx/filefn.h>
#include <filezilla.h>
#include "queue.h"
#include "Mainfrm.h"
#include "Options.h"
#include <libfilezilla/local_filesys.hpp>




std::string CryptFiles::getFileNameByType(CryptFileType cfType)
{
	switch (cfType)
	{
	case CryptFileType::FileZillaXml:
		return "filezilla.dat";

	case CryptFileType::SiteManagerXml:
		return "sitemanager.dat";

	case CryptFileType::RecentServersXml:
		return "recentservers.dat";

	case CryptFileType::QueueSqlite3:
		return "queueDB.dat";

	case CryptFileType::QueueXml:
		return "queue.dat";

	case CryptFileType::BookmarksXml:
		return "bookmarks.dat";

	case CryptFileType::TrustedcertsXml:
		return "trustedcerts.dat";
	}

	return "";
}

AesFile::KeyFile & CryptFiles::getKeyFile(CryptFileType cfType)
{
	switch (cfType)
	{
	case CryptFileType::FileZillaXml:
		return keyFileZillaXml;

	case CryptFileType::SiteManagerXml:
		return keySiteManagerXml;

	case CryptFileType::RecentServersXml:
		return keyRecentServersXml;

	case CryptFileType::QueueSqlite3:
		return keyQueueSqlite3;

	case CryptFileType::QueueXml:
		return keyQueueXml;

	case CryptFileType::BookmarksXml:
		return keyBookmarksXml;

	case CryptFileType::TrustedcertsXml:
		return keyTrustedcertsXml;
	}

	return keyFileNull;
}

void CryptFiles::generateNewKeys(const wxString & pass)
{
	ShaMasterPass mpass(pass);

	keyFileNull = mpass.generateKeyForFile("NULL");

	generateKeyFile(mpass, CryptFileType::FileZillaXml);
	generateKeyFile(mpass, CryptFileType::SiteManagerXml);
	generateKeyFile(mpass, CryptFileType::RecentServersXml);
	generateKeyFile(mpass, CryptFileType::QueueSqlite3);
	generateKeyFile(mpass, CryptFileType::QueueXml);
	generateKeyFile(mpass, CryptFileType::BookmarksXml);
	generateKeyFile(mpass, CryptFileType::TrustedcertsXml);
}

void CryptFiles::setNewPassword(const wxString & pass)
{
	generateNewKeys(pass);
}

bool CryptFiles::checkCryptedFile(CryptFileType cfType)
{
	std::vector<uint8_t> cryptedData;
	if (loadAllDataFromFile(cfType, cryptedData))
	{
		AesFile::KeyFile keyFile = getKeyFile(cfType);
		AesFile aesFile(keyFile);
		std::vector<uint8_t> plainData;
		if (aesFile.decryptFileData(cryptedData, plainData))
			return true;
	}
	return false;
}

bool CryptFiles::loadAllDataFromFile(CryptFileType cfType, std::vector<uint8_t>& data)
{
	data.clear();

	wxLogNull logNo;

	wxString filePath;
	if (CryptFiles::getFilePathByType(cfType, filePath))
	{
		if (CryptFiles::isFileExists(filePath))
		{
			if (wxFile::Access(filePath, wxFile::read))
			{
				wxFile file(filePath, wxFile::read);
				if (file.IsOpened())
				{
					size_t file_size = (size_t)file.Length();
					data.resize(file_size, 0);
					size_t real_real = (size_t)file.Read(data.data(), data.size());
					file.Close();

					if (real_real == data.size())
						return true;
				}
			}
		}
	}

	data.clear();
	return false;
}

bool CryptFiles::saveAllDataToFile(CryptFileType cfType, const std::vector<uint8_t>& data)
{
	wxLogNull logNo;

	wxString filePath;
	if (CryptFiles::getFilePathByType(cfType, filePath))
	{
		wxFile file(filePath, wxFile::write);
		if (file.IsOpened())
		{
			size_t real_write = (size_t)file.Write(data.data(), data.size());
			file.Close();

			if (real_write == data.size())
				return true;
		}
	}

	return false;
}

std::vector<uint8_t> CryptFiles::encryptFileData(const std::vector<uint8_t>& plainData, CryptFileType cfType)
{	
	AesFile::KeyFile keyFile = getKeyFile(cfType);

	AesFile aesFile(keyFile);
	std::vector<uint8_t> cryptedData = aesFile.encryptFileData(plainData);
	return cryptedData;
}

bool CryptFiles::decryptFileData(const std::vector<uint8_t>& cryptedData, std::vector<uint8_t>& plainData, CryptFileType cfType)
{
	AesFile::KeyFile keyFile = getKeyFile(cfType);

	AesFile aesFile(keyFile);
	return aesFile.decryptFileData(cryptedData, plainData);
}

void CryptFiles::generateKeyFile(ShaMasterPass & mpass, CryptFileType cfType)
{
	AesFile::KeyFile &keyFile = getKeyFile(cfType);
	keyFile = mpass.generateKeyForFile(getFileNameByType(cfType));
}

bool CryptFiles::cryptFilesExist()
{
	wxString fpath;
	if (CryptFiles::getFilePathByType(CryptFileType::FileZillaXml, fpath))
	{
		if (isFileExists(fpath))
			return true;
	}

	return false;
}

bool CryptFiles::cryptFilesCheckMasterPass(const wxString & masterPass)
{
	wxString fpath;
	if (CryptFiles::getFilePathByType(CryptFileType::FileZillaXml, fpath))
	{
		if (isFileExists(fpath))
		{
			setNewPassword(masterPass);
			if (checkCryptedFile(CryptFileType::FileZillaXml))
				return true;
		}
	}

	return false;
}

void CryptFiles::cryptFilesCreate(const wxString & masterPass)
{
	setNewPassword(masterPass);

	CryptFiles::deleteAllCryptedFiles();
	clearCryptedFile(CryptFileType::FileZillaXml);
}

void CryptFiles::cryptFilesChangeMasterPass(const wxString & oldMasterPass, const wxString & newMasterPass)
{
	setNewPassword(newMasterPass);

	changePassOnFile(CryptFileType::FileZillaXml, oldMasterPass, newMasterPass);
	changePassOnFile(CryptFileType::SiteManagerXml, oldMasterPass, newMasterPass);
	changePassOnFile(CryptFileType::RecentServersXml, oldMasterPass, newMasterPass);
	changePassOnFile(CryptFileType::QueueSqlite3, oldMasterPass, newMasterPass);
	changePassOnFile(CryptFileType::QueueXml, oldMasterPass, newMasterPass);
	changePassOnFile(CryptFileType::BookmarksXml, oldMasterPass, newMasterPass);
	changePassOnFile(CryptFileType::TrustedcertsXml, oldMasterPass, newMasterPass);

	setNewPassword(newMasterPass);
}

bool CryptFiles::getFilePathByType(CryptFileType cfType, wxString & filePath)
{
	filePath.Clear();

	const wxString fname = wxString(getFileNameByType(cfType));
	if (!fname.IsEmpty())
	{
		const wxString dir_settings = wxString(COptions::GetUnadjustedSettingsDir().GetPath());
		filePath = getRedirectedName(dir_settings + fname);
		return true;
	}
	
	return false;
}

bool CryptFiles::isFileExists(const wxString & filePath)
{
	wxLogNull logNo;

	const int pos = filePath.Find('*');
	if (pos < 0)
		return wxFileExists(filePath);

	wxASSERT(pos > 0);
	wxASSERT(filePath[pos - 1] == '/');
	wxASSERT(filePath.size() > static_cast<size_t>(pos + 1) && filePath[pos + 1] == '/');

	wxDir dir(filePath.Left(pos));
	if (!dir.IsOpened())
		return false;

	wxString subDir;
	bool found = dir.GetFirst(&subDir, _T(""), wxDIR_DIRS);
	while (found) 
	{
		if (isFileExists(filePath.Left(pos) + subDir + filePath.Mid(pos + 1)))
			return true;

		found = dir.GetNext(&subDir);
	}

	return false;
}

void CryptFiles::deleteCryptedFile(CryptFileType cfType)
{
	wxLogNull logNo;

	wxString fpath;
	if (CryptFiles::getFilePathByType(cfType, fpath))
	{
		if (isFileExists(fpath))
			wxRemoveFile(fpath);
	}
}

void CryptFiles::deleteAllCryptedFiles()
{
	CryptFiles::deleteCryptedFile(CryptFileType::FileZillaXml);
	CryptFiles::deleteCryptedFile(CryptFileType::SiteManagerXml);
	CryptFiles::deleteCryptedFile(CryptFileType::RecentServersXml);
	CryptFiles::deleteCryptedFile(CryptFileType::QueueSqlite3);
	CryptFiles::deleteCryptedFile(CryptFileType::QueueXml);
	CryptFiles::deleteCryptedFile(CryptFileType::BookmarksXml);
	CryptFiles::deleteCryptedFile(CryptFileType::TrustedcertsXml);
}

void CryptFiles::clearCryptedFile(CryptFileType cfType)
{
	wxString fpath;
	if (CryptFiles::getFilePathByType(cfType, fpath))
	{
		const std::vector<uint8_t> plain;
		const std::vector<uint8_t> crypted_data = CryptFiles::encryptFileData(plain, cfType);
			
		saveAllDataToFile(cfType, crypted_data);
	}
}

void CryptFiles::changePassOnFile(CryptFileType cfType, const wxString &oldMasterPass, const wxString &newMasterPass)
{
	wxString fpath;
	if (CryptFiles::getFilePathByType(cfType, fpath))
	{
		if (CryptFiles::isFileExists(fpath))
		{
			setNewPassword(oldMasterPass);
			std::vector<uint8_t> oldPlainData;
			{
				std::vector<uint8_t> oldCryptedData;
				if (CryptFiles::loadAllDataFromFile(cfType, oldCryptedData))
				{
					if (!decryptFileData(oldCryptedData, oldPlainData, cfType))
						oldPlainData.clear();
				}
			}

			deleteCryptedFile(cfType);

			setNewPassword(newMasterPass);
			const std::vector<uint8_t> newCryptedData = encryptFileData(oldPlainData, cfType);
			CryptFiles::saveAllDataToFile(cfType, newCryptedData);
		}
	}
}

wxString CryptFiles::getRedirectedName(const wxString &filePath)
{
	wxString redirectedName = filePath;
	
	bool isLink = false;
	if (fz::local_filesys::get_file_info(fz::to_native(redirectedName), isLink, 0, 0, 0) == fz::local_filesys::file) 
	{
		if (isLink) 
		{
			CLocalPath target(fz::local_filesys::get_link_target(fz::to_native(redirectedName)));
			if (!target.empty()) 
			{
				redirectedName = target.GetPath();
				redirectedName.RemoveLast();
			}
		}
	}
	
	return redirectedName;
}
