#include "cryqueue_storage_impl.h"
#include "cryqueue_storage.h"
#include <filezilla.h>
#include "queue.h"
#include "Mainfrm.h"
#include "Options.h"
#include <wx/wx.h>
#include <unordered_map>
#include "Common.h"
#include "CryptFiles.h"
#include "ServerElement.h"


#define INVALID_DATA -1


CCryptedQueueStorageImpl::CCryptedQueueStorageImpl()
{
}

CCryptedQueueStorageImpl::~CCryptedQueueStorageImpl()
{
}

bool CCryptedQueueStorageImpl::saveQueue(std::vector<CServerItem*> const& queue)
{
	CryptedQueueDataBase cDB(CryptFileType::QueueSqlite3);
	if (cDB.loadDBFromCryptedFileAndClearFile())
	{
		if (saveQueueToDB(queue, cDB))
		{
			if (cDB.saveDBToCryptedFile())
				return true;
		}
	}

	return false;
}

bool CCryptedQueueStorageImpl::loadQueue(std::vector<ServerElement> &queue)
{
	CryptedQueueDataBase cDB(CryptFileType::QueueSqlite3);
	if (cDB.loadDBFromCryptedFileAndClearFile())
	{
		if (loadQueueFromDB(queue, cDB))
			return true;
	}

	return false;
}

bool CCryptedQueueStorageImpl::saveQueueToDB(std::vector<CServerItem*> const & queue, CryptedQueueDataBase & cDB)
{
	ClearCaches();

	bool ret = true;

	for (std::vector<CServerItem*>::const_iterator it = queue.begin(); it != queue.end(); ++it)
		ret &= SaveServer(**it, cDB);

	ClearCaches();

	return ret;
}

bool CCryptedQueueStorageImpl::loadQueueFromDB(std::vector<ServerElement>& queue, CryptedQueueDataBase & cDB)
{
	queue.clear();

	bool error = false;

	ReadLocalPaths(cDB);
	ReadRemotePaths(cDB);

	std::vector<CIDLineServer> lines_servers = cDB.tableServersReadAllLines();
	for (const CIDLineServer &idLine : lines_servers)
	{
		const int64_t server_id = idLine.first;
		const CDBTableServerLine &serverLine = idLine.second;

		CServer server;
		if ((server_id > 0) && (ParseServerFromRow(server, serverLine) > 0))
		{
			ServerElement server_element;
			server_element.server = server;
			server_element.files.clear();

			std::vector<CIDLineFile> file_lines = cDB.tableFilesReadAllLinesByServerId(server_id);
			for (const CIDLineFile &idLineFile : file_lines)
			{
				const int64_t file_id = idLineFile.first;
				const CDBTableFileLine &fileLine = idLineFile.second;

				if (file_id > 0)
				{
					CFileItem* fileItem = 0;
					if (ParseFileFromRow(&fileItem, fileLine) > 0 && fileItem != 0)
						server_element.files.push_back(fileItem);
				}
			}

			queue.push_back(server_element);
		}
	}

	ClearCaches();

	return !error;
}

void CCryptedQueueStorageImpl::ReadLocalPaths(CryptedQueueDataBase & cDB)
{
	std::vector<CIDLinePath> localLines = cDB.tableLocalPathsReadAllLines();
	for (const CIDLinePath &idLine : localLines)
	{
		const int64_t id = idLine.first;//GetColumnInt64(selectLocalPathQuery_, path_table_column_names::id);

		const CDBTablePathLine &pathLine = idLine.second;
		const wxString localPathRaw = pathLine.path.getValue();//GetColumnText(selectLocalPathQuery_, path_table_column_names::path);

		CLocalPath localPath;
		if (id > 0 && !localPathRaw.empty() && localPath.SetPath(localPathRaw.ToStdWstring()))
			reverseLocalPaths_[id] = localPath;
	}
}

void CCryptedQueueStorageImpl::ReadRemotePaths(CryptedQueueDataBase & cDB)
{
	std::vector<CIDLinePath> remoteLines = cDB.tableRemotePathsReadAllLines();
	for (const CIDLinePath &idLine : remoteLines)
	{
		const int64_t id = idLine.first;//int64_t id = GetColumnInt64(selectRemotePathQuery_, path_table_column_names::id);

		const CDBTablePathLine &pathLine = idLine.second;
		const wxString remotePathRaw = pathLine.path.getValue();//wxString remotePathRaw = GetColumnText(selectRemotePathQuery_, path_table_column_names::path);

		CServerPath remotePath;
		if (id > 0 && !remotePathRaw.empty() && remotePath.SetSafePath(remotePathRaw))
			reverseRemotePaths_[id] = remotePath;
	}
}

const CLocalPath& CCryptedQueueStorageImpl::GetLocalPath(int64_t id) const
{
	std::map<int64_t, CLocalPath>::const_iterator it = reverseLocalPaths_.find(id);
	if (it != reverseLocalPaths_.end())
		return it->second;

	static CLocalPath const empty{};
	return empty;
}

const CServerPath& CCryptedQueueStorageImpl::GetRemotePath(int64_t id) const
{
	std::map<int64_t, CServerPath>::const_iterator it = reverseRemotePaths_.find(id);
	if (it != reverseRemotePaths_.end())
		return it->second;

	static CServerPath const empty{};
	return empty;
}

void CCryptedQueueStorageImpl::ClearCaches()
{
	localPaths_.clear();
	remotePaths_.clear();
	reverseLocalPaths_.clear();
	reverseRemotePaths_.clear();
}

int64_t CCryptedQueueStorageImpl::SaveLocalPath(const CLocalPath& path, CryptedQueueDataBase & cDB)
{
	const wxString p = wxString(path.GetPath());

	std::unordered_map<wxString, int64_t, wxStringHash, fast_equal>::const_iterator it = localPaths_.find(p);
	if (it != localPaths_.end())
		return it->second;

	CDBTablePathLine pathLine;
	pathLine.path.setValue(p);

	const int64_t id = cDB.tableLocalPathsAddLine(pathLine);
	
	if (id > 0)
	{
		localPaths_[p] = id;
		return id;
	}

	return -1;
}

int64_t CCryptedQueueStorageImpl::SaveRemotePath(const CServerPath& path, CryptedQueueDataBase & cDB)
{
	const wxString safePath = path.GetSafePath();
	
	std::unordered_map<wxString, int64_t, wxStringHash>::const_iterator it = remotePaths_.find(safePath);
	if (it != remotePaths_.end())
		return it->second;

	CDBTablePathLine pathLine;
	pathLine.path.setValue(safePath);

	const int64_t id = cDB.tableRemotePathsAddLine(pathLine);
	
	if (id > 0)
	{
		remotePaths_[safePath] = id;
		return id;
	}

	return -1;
}

bool CCryptedQueueStorageImpl::SaveFile(const CFileItem& file, int64_t server_id, CryptedQueueDataBase & cDB)
{
	CDBTableFileLine fileLine;

	fileLine.server_id.setValue(server_id);

	if (file.m_edit != CEditHandler::none)
		return true;

	fileLine.source_file.setValue(wxString(file.GetSourceFile()));

	auto const& targetFile = file.GetTargetFile();
	if (targetFile)
	{
		const wxString str_target = wxString(*targetFile);
		fileLine.target_file.setValue(str_target);
	}
	else
		fileLine.target_file.setNull();

	int64_t localPathId = SaveLocalPath(file.GetLocalPath(), cDB);
	int64_t remotePathId = SaveRemotePath(file.GetRemotePath(), cDB);
	if (localPathId == -1 || remotePathId == -1)
		return false;

	fileLine.local_path_id.setValue(localPathId);
	fileLine.remote_path_id.setValue(remotePathId);
	
	fileLine.download.setValue(file.Download() ? 1 : 0);
		
	if (file.GetSize() != -1)
		fileLine.size.setValue(file.GetSize());
	else
		fileLine.size.setNull();
	
	if (file.m_errorCount)
		fileLine.error_count.setValue(static_cast<uint64_t>(file.m_errorCount));
	else
		fileLine.error_count.setNull();
	
	fileLine.priority.setValue(static_cast<int64_t>(file.GetPriority()));
	
	fileLine.ascii_file.setValue(file.Ascii() ? 1 : 0);
	
	if (file.m_defaultFileExistsAction != CFileExistsNotification::unknown)
		fileLine.default_exists_action.setValue(static_cast<int64_t>(file.m_defaultFileExistsAction));
	else
		fileLine.default_exists_action.setNull();
		
	int64_t file_id = cDB.tableFilesAddLine(fileLine);

	if (file_id > 0)
		return true;

	return false;
}

bool CCryptedQueueStorageImpl::SaveDirectory(const CFolderItem& directory, int64_t server_id, CryptedQueueDataBase & cDB)
{
	CDBTableFileLine fileLine;

	fileLine.server_id.setValue(server_id);
	
	if (directory.Download())
		fileLine.source_file.setNull();
	else
		fileLine.source_file.setValue(wxString(directory.GetSourceFile()));

	fileLine.target_file.setNull();

	int64_t localPathId = directory.Download() ? SaveLocalPath(directory.GetLocalPath(), cDB) : -1;
	int64_t remotePathId = directory.Download() ? -1 : SaveRemotePath(directory.GetRemotePath(), cDB);
	if (localPathId == -1 && remotePathId == -1)
		return false;

	fileLine.local_path_id.setValue(localPathId);
	fileLine.remote_path_id.setValue(remotePathId);
	
	fileLine.download.setValue(directory.Download() ? 1 : 0);
	
	fileLine.size.setNull();
	
	if (directory.m_errorCount)
		fileLine.error_count.setValue(static_cast<uint64_t>(directory.m_errorCount));
	else
		fileLine.error_count.setNull();
	
	fileLine.priority.setValue(static_cast<int64_t>(directory.GetPriority()));

	fileLine.ascii_file.setNull();

	fileLine.default_exists_action.setNull();
	
	int64_t file_id = cDB.tableFilesAddLine(fileLine);

	if (file_id > 0)
		return true;

	return false;
}

bool CCryptedQueueStorageImpl::SaveServer(const CServerItem& item, CryptedQueueDataBase& cDB)
{
	const CServer& server = item.GetServer();

	int64_t server_id = 0;
	bool ret = serverTableAddServer(server, server_id, cDB);

	if (ret && server_id > 0)
	{
		const std::vector<CQueueItem*>& children = item.GetChildren();
		for (std::vector<CQueueItem*>::const_iterator it = children.begin() + item.GetRemovedAtFront(); it != children.end(); ++it)
		{
			CQueueItem & childItem = **it;
			if (childItem.GetType() == QueueItemType::File)
				ret &= SaveFile(static_cast<CFileItem&>(childItem), server_id, cDB);
			else if (childItem.GetType() == QueueItemType::Folder)
				ret &= SaveDirectory(static_cast<CFolderItem&>(childItem), server_id, cDB);
		}
	}

	return ret;
}

bool CCryptedQueueStorageImpl::serverTableAddServer(const CServer& server, int64_t &server_id, CryptedQueueDataBase & cDB)
{
	server_id = 0;
	const bool kiosk_mode = COptions::Get()->GetOptionVal(OPTION_DEFAULT_KIOSKMODE) != 0;

	CDBTableServerLine serverLine;

	serverLine.host.setValue(wxString(server.GetHost()));
	serverLine.port.setValue(static_cast<uint64_t>(server.GetPort()));
	serverLine.protocol.setValue(static_cast<int64_t>(server.GetProtocol()));
	serverLine.type.setValue(static_cast<int64_t>(server.GetType()));

	enum LogonType logonType = server.GetLogonType();
	if (server.GetLogonType() != ANONYMOUS)
	{
		serverLine.user.setValue(wxString(server.GetUser()));

		if (server.GetLogonType() == NORMAL || server.GetLogonType() == ACCOUNT)
		{
			if (kiosk_mode)
			{
				logonType = ASK;

				serverLine.password.setNull();
				serverLine.account.setNull();
			}
			else
			{
				serverLine.password.setValue(wxString(server.GetPass()));
				
				if (server.GetLogonType() == ACCOUNT)
					serverLine.account.setValue(wxString(server.GetAccount()));
				else
					serverLine.account.setNull();
			}
		}
		else
		{
			serverLine.password.setNull();
			serverLine.account.setNull();
		}
	}
	else
	{
		serverLine.user.setNull();
		serverLine.password.setNull();
		serverLine.account.setNull();
	}

	serverLine.logontype.setValue(static_cast<int64_t>(logonType));

	serverLine.timezone_offset.setValue(static_cast<int64_t>(server.GetTimezoneOffset()));
	
	switch (server.GetPasvMode())
	{
	case MODE_PASSIVE:
		serverLine.transfer_mode.setValue(wxString(_T("passive")));
		break;
	
	case MODE_ACTIVE:
		serverLine.transfer_mode.setValue(wxString(_T("active")));
		break;
	
	default:
		serverLine.transfer_mode.setValue(wxString(_T("default")));
		break;
	}

	serverLine.max_connections.setValue(static_cast<int64_t>(server.MaximumMultipleConnections()));

	switch (server.GetEncodingType())
	{
	default:
	case ENCODING_AUTO:
		serverLine.encoding.setValue(wxString(_T("Auto")));
		break;

	case ENCODING_UTF8:
		serverLine.encoding.setValue(wxString(_T("UTF-8")));
		break;

	case ENCODING_CUSTOM:
		serverLine.encoding.setValue(wxString(server.GetCustomEncoding()));
		break;
	}


	if (CServer::SupportsPostLoginCommands(server.GetProtocol()))
	{
		const std::vector<wxString>& postLoginCommands = server.GetPostLoginCommands();
		if (!postLoginCommands.empty())
		{
			wxString commands;
			for (std::vector<wxString>::const_iterator iter = postLoginCommands.begin(); iter != postLoginCommands.end(); ++iter)
			{
				if (!commands.empty())
					commands += _T("\n");
				commands += *iter;
			}
			serverLine.post_login_commands.setValue(commands);
		}
		else
			serverLine.post_login_commands.setNull();
	}
	else
		serverLine.post_login_commands.setNull();


	serverLine.bypass_proxy.setValue(server.GetBypassProxy() ? 1 : 0);
	
	if (!server.GetName().empty())
		serverLine.name.setValue(wxString(server.GetName()));
	else
		serverLine.name.setNull();


	const int64_t id = cDB.tableServersAddLine(serverLine);

	if (id > 0)
	{
		server_id = id;
		return true;
	}

	return false;
}

int64_t CCryptedQueueStorageImpl::ParseServerFromRow(CServer& server, const CDBTableServerLine &serverLine)
{
	server = CServer();

	const wxString host = serverLine.host.getValue();//GetColumnText(selectServersQuery_, server_table_column_names::host);
	if (host.empty())
		return INVALID_DATA;

	int port = (int)serverLine.port.getValue();//GetColumnInt(selectServersQuery_, server_table_column_names::port);
	if (port < 1 || port > 65535)
		return INVALID_DATA;

	if (!server.SetHost(host, port))
		return INVALID_DATA;

	const int protocol = (int)serverLine.protocol.getValue();//GetColumnInt(selectServersQuery_, server_table_column_names::protocol);
	if (protocol < 0 || protocol > MAX_VALUE)
		return INVALID_DATA;
	server.SetProtocol(static_cast<ServerProtocol>(protocol));

	int type = (int)serverLine.type.getValue();//GetColumnInt(selectServersQuery_, server_table_column_names::type);
	if (type < 0 || type >= SERVERTYPE_MAX)
		return INVALID_DATA;

	server.SetType((enum ServerType)type);

	int logonType = (int)serverLine.logontype.getValue();//GetColumnInt(selectServersQuery_, server_table_column_names::logontype);
	if (logonType < 0 || logonType >= LOGONTYPE_MAX)
		return INVALID_DATA;

	server.SetLogonType((enum LogonType)logonType);

	if (server.GetLogonType() != ANONYMOUS)
	{
		const wxString user = serverLine.user.getValue();//GetColumnText(selectServersQuery_, server_table_column_names::user);

		wxString pass;
		if ((long)NORMAL == logonType || (long)ACCOUNT == logonType)
			pass = serverLine.password.getValue();//GetColumnText(selectServersQuery_, server_table_column_names::password);

		if (!server.SetUser(user, pass))
			return INVALID_DATA;

		if ((long)ACCOUNT == logonType)
		{
			const wxString account = serverLine.account.getValue();//GetColumnText(selectServersQuery_, server_table_column_names::account);
			if (account.empty())
				return INVALID_DATA;
			if (!server.SetAccount(account))
				return INVALID_DATA;
		}
	}

	const int timezoneOffset = (int)serverLine.timezone_offset.getValue();//GetColumnInt(selectServersQuery_, server_table_column_names::timezone_offset);
	if (!server.SetTimezoneOffset(timezoneOffset))
		return INVALID_DATA;

	const wxString pasvMode = serverLine.transfer_mode.getValue();//GetColumnText(selectServersQuery_, server_table_column_names::transfer_mode);
	if (pasvMode == _T("passive"))
		server.SetPasvMode(MODE_PASSIVE);
	else if (pasvMode == _T("active"))
		server.SetPasvMode(MODE_ACTIVE);
	else
		server.SetPasvMode(MODE_DEFAULT);

	const int maximumMultipleConnections = (int)serverLine.max_connections.getValue();//GetColumnInt(selectServersQuery_, server_table_column_names::max_connections);
	if (maximumMultipleConnections < 0)
		return INVALID_DATA;
	server.MaximumMultipleConnections(maximumMultipleConnections);

	const wxString encodingType = serverLine.encoding.getValue();//GetColumnText(selectServersQuery_, server_table_column_names::encoding);
	if (encodingType.empty() || encodingType == _T("Auto"))
		server.SetEncodingType(ENCODING_AUTO);
	else if (encodingType == _T("UTF-8"))
		server.SetEncodingType(ENCODING_UTF8);
	else
	{
		if (!server.SetEncodingType(ENCODING_CUSTOM, encodingType))
			return INVALID_DATA;
	}


	if (CServer::SupportsPostLoginCommands(server.GetProtocol()))
	{
		std::vector<wxString> postLoginCommands;

		wxString commands = serverLine.post_login_commands.getValue();//GetColumnText(selectServersQuery_, server_table_column_names::post_login_commands);
		while (!commands.empty())
		{
			int pos = commands.Find('\n');
			if (!pos)
				commands = commands.Mid(1);
			else if (pos == -1)
			{
				postLoginCommands.push_back(commands);
				commands.clear();
			}
			else
			{
				postLoginCommands.push_back(commands.Left(pos));
				commands = commands.Mid(pos + 1);
			}
		}
		if (!server.SetPostLoginCommands(postLoginCommands))
			return INVALID_DATA;
	}

	const int64_t ibypass_proxy = serverLine.bypass_proxy.getValue();//GetColumnInt(selectServersQuery_, server_table_column_names::bypass_proxy);
	server.SetBypassProxy(ibypass_proxy == 1);
	
	const wxString tname = serverLine.name.getValue();//GetColumnText(selectServersQuery_, server_table_column_names::name);
	server.SetName(tname);

	return 1;
}

int64_t CCryptedQueueStorageImpl::ParseFileFromRow(CFileItem** pItem, const CDBTableFileLine &fileLine)
{
	*pItem = 0;

	wxString sourceFile = fileLine.source_file.getValue();//GetColumnText(selectFilesQuery_, file_table_column_names::source_file);
	
	wxString targetFile = fileLine.target_file.getValue();//GetColumnText(selectFilesQuery_, file_table_column_names::target_file);

	int64_t localPathId = fileLine.local_path_id.getValue();//GetColumnInt64(selectFilesQuery_, file_table_column_names::local_path);
	int64_t remotePathId = fileLine.remote_path_id.getValue();//GetColumnInt64(selectFilesQuery_, file_table_column_names::remote_path);

	const CLocalPath& localPath(GetLocalPath(localPathId));
	const CServerPath& remotePath(GetRemotePath(remotePathId));

	int64_t idownload = fileLine.download.getValue();//GetColumnInt64(selectFilesQuery_, file_table_column_names::download);

	bool download = idownload != 0;

	if (localPathId == -1 || remotePathId == -1)
	{
		// QueueItemType::Folder
		if ((download && localPath.empty()) ||
			(!download && remotePath.empty()))
		{
			return INVALID_DATA;
		}

		if (download)
			*pItem = new CFolderItem(0, true, localPath);
		else
			*pItem = new CFolderItem(0, true, remotePath, sourceFile);
	}
	else
	{
		const int64_t size = fileLine.size.getValue();//GetColumnInt64(selectFilesQuery_, file_table_column_names::size);
		
		const int64_t ierror_count = fileLine.error_count.getValue();//GetColumnInt64(selectFilesQuery_, file_table_column_names::error_count)
		const unsigned char errorCount = static_cast<unsigned char>(static_cast<uint64_t>(ierror_count));
		
		const int priority = (int)fileLine.priority.getValue(static_cast<int64_t>(QueuePriority::normal));//GetColumnInt64(selectFilesQuery_, file_table_column_names::priority, static_cast<int>(QueuePriority::normal));

		const int64_t iascii_file = fileLine.ascii_file.getValue();//GetColumnInt64(selectFilesQuery_, file_table_column_names::ascii_file);
		const bool ascii = iascii_file != 0;

		int overwrite_action = (int)fileLine.default_exists_action.getValue(static_cast<int64_t>(CFileExistsNotification::unknown));//GetColumnInt64(selectFilesQuery_, file_table_column_names::default_exists_action, CFileExistsNotification::unknown);

		if (sourceFile.empty() || localPath.empty() ||
			remotePath.empty() ||
			size < -1 ||
			priority < 0 || priority >= static_cast<int>(QueuePriority::count))
		{
			return INVALID_DATA;
		}

		CFileItem* fileItem = new CFileItem(0, true, download, sourceFile, targetFile, localPath, remotePath, size);
		*pItem = fileItem;
		fileItem->SetAscii(ascii);
		fileItem->SetPriorityRaw(QueuePriority(priority));
		fileItem->m_errorCount = errorCount;

		if (overwrite_action > 0 && overwrite_action < CFileExistsNotification::ACTION_COUNT)
			fileItem->m_defaultFileExistsAction = (CFileExistsNotification::OverwriteAction)overwrite_action;
	}

	return 1;
}
