#include <filezilla.h>
#include "buildinfo.h"
#include "xmlfunctions.h"
#include "cryxmlfunctions.h"
#include "Options.h"
#include <wx/ffile.h>
#include <wx/log.h>
#include <wx/base64.h>

#include <libfilezilla/local_filesys.hpp>

#include <vector>


CCryptedXmlFile::CCryptedXmlFile(CryptFileType cfType, wxString const& root)
	: mCryptedFileType(cfType)
{
	if (!root.empty()) {
		m_rootName = root;
	}

	wxString fpath;
	if (CryptFiles::getFilePathByType(cfType, fpath))
		SetFileName(fpath);
}

void CCryptedXmlFile::SetFileName(const wxString& name)
{
	wxASSERT(!name.empty());
	m_fileName = name;
	m_modificationTime = fz::datetime();
}

CCryptedXmlFile::~CCryptedXmlFile()
{
	Close();
}

pugi::xml_node CCryptedXmlFile::Load()
{
	Close();
	m_error.clear();

	wxCHECK(!m_fileName.empty(), m_element);

	wxString redirectedName;
	CryptFiles::getFilePathByType(mCryptedFileType, redirectedName);
		

	GetXmlFile(redirectedName);
	if (!m_element) 
	{
		// Loading failed. If file is empty, create new file.
		m_error.clear();
		CreateEmpty();

		m_modificationTime = fz::local_filesys::get_modification_time(fz::to_native(redirectedName));
		return m_element;
	}

	m_modificationTime = fz::local_filesys::get_modification_time(fz::to_native(redirectedName));
	return m_element;
}

bool CCryptedXmlFile::Modified()
{
	wxCHECK(!m_fileName.empty(), false);

	if (m_modificationTime.empty())
		return true;

	fz::datetime const modificationTime = fz::local_filesys::get_modification_time(fz::to_native(m_fileName));
	if (!modificationTime.empty() && modificationTime == m_modificationTime)
		return false;

	return true;
}

void CCryptedXmlFile::Close()
{
	m_element = pugi::xml_node();
	m_document.reset();
}

void CCryptedXmlFile::UpdateMetadata()
{
	if (!m_element || std::string(m_element.name()) != "FileZilla3") {
		return;
	}

	SetTextAttribute(m_element, "version", CBuildInfo::GetVersion());

	wxString const platform =
#ifdef __WXMSW__
		_T("windows");
#elif defined(__WXMAX__)
		_T("mac");
#else
		_T("*nix");
#endif
	SetTextAttribute(m_element, "platform", platform);
}

bool CCryptedXmlFile::Save(bool printError)
{
	m_error.clear();

	wxCHECK(!m_fileName.empty(), false);
	wxCHECK(m_document, false);

	UpdateMetadata();

	bool res = SaveXmlFile();
	m_modificationTime = fz::local_filesys::get_modification_time(fz::to_native(m_fileName));

	if (!res && printError) {
		wxASSERT(!m_error.empty());

		wxString msg = wxString::Format(_("Could not write \"%s\":"), m_fileName);
		wxMessageBoxEx(msg + _T("\n") + m_error, _("Error writing xml file"), wxICON_ERROR);
	}
	return res;
}

pugi::xml_node CCryptedXmlFile::CreateEmpty()
{
	Close();

	pugi::xml_node decl = m_document.append_child(pugi::node_declaration);
	decl.append_attribute("version") = "1.0";
	decl.append_attribute("encoding") = "UTF-8";

	m_element = m_document.append_child(m_rootName.utf8_str());
	return m_element;
}


// Opens the specified XML file if it exists or creates a new one otherwise.
// Returns 0 on error.
bool CCryptedXmlFile::GetXmlFile(wxString const& file)
{
	Close();

	if (fz::local_filesys::get_size(fz::to_native(file)) <= 0) {
		return false;
	}

	// File exists, open it
	//auto result = m_document.load_file(static_cast<wchar_t const*>(file.c_str()));
	// Load document from buffer. Copies/converts the buffer, so it may be deleted or changed after the function returns.
	//xml_parse_result load_buffer(const void* contents, size_t size, unsigned int options = parse_default, xml_encoding encoding = encoding_auto);
	
	std::vector<uint8_t> cryptedData;
	if (!CryptFiles::loadAllDataFromFile(mCryptedFileType, cryptedData))
	{
		m_error += wxString(_T("Can't load data from file ")) + file;
		return false;
	}

	std::vector<uint8_t> plainData;
	if (!CryptFiles::instance().decryptFileData(cryptedData, plainData, mCryptedFileType))
	{
		m_error += wxString(_T("Can't parse data in file ")) + file;
		return false;
	}

	auto result = m_document.load_buffer(plainData.data(), plainData.size());

	//
	//load_buffer(const void* contents, size_t size);
	
	if (!result) {
		m_error += wxString::Format(_T("%s at offset %lld."), wxString(result.description()), static_cast<long long>(result.offset));
		return false;
	}

	m_element = m_document.child(m_rootName.utf8_str());
	if (!m_element) {
		if (m_document.first_child()) { // Beware: parse_declaration and parse_doctype can break this
			// Not created by FileZilla3
			Close();
			m_error = _("Unknown root element, the file does not appear to be generated by FileZilla.");
			return false;
		}
		m_element = m_document.append_child(m_rootName.utf8_str());
	}

	return true;
}



bool CCryptedXmlFile::SaveXmlFile()
{
	bool isLink = false;
	int flags = 0;

	wxString redirectedName;
	CryptFiles::getFilePathByType(mCryptedFileType, redirectedName);

	if (fz::local_filesys::get_file_info(fz::to_native(redirectedName), isLink, 0, 0, &flags) == fz::local_filesys::file) 
	{
#ifdef __WXMSW__
		if (flags & FILE_ATTRIBUTE_HIDDEN)
			SetFileAttributes(redirectedName.c_str(), flags & ~FILE_ATTRIBUTE_HIDDEN);
#endif
	}

	const size_t xml_file_size = GetRawDataLength();
	std::vector<uint8_t> xml_file_data(xml_file_size, 0);
	GetRawDataHere((char*)xml_file_data.data(), xml_file_data.size());

	std::vector<uint8_t> crypted_file_data = CryptFiles::instance().encryptFileData(xml_file_data, mCryptedFileType);
	bool success = CryptFiles::saveAllDataToFile(mCryptedFileType, crypted_file_data);

	if (!success) 
	{
		wxRemoveFile(redirectedName);
		m_error = _("Failed to write xml file");
		return false;
	}

	return true;
}


namespace {
struct xml_memory_writer : pugi::xml_writer
{
	size_t written{};
	char* buffer{};
	size_t remaining{};

	virtual void write(const void* data, size_t size)
	{
		if (buffer && size <= remaining) {
			memcpy(buffer, data, size);
			buffer += size;
			remaining -= size;
		}
		written += size;
	}
};
}

size_t CCryptedXmlFile::GetRawDataLength()
{
	if (!m_document)
		return 0;

	xml_memory_writer writer;
	m_document.save(writer);
	return writer.written;
}

void CCryptedXmlFile::GetRawDataHere(char* p, size_t size) // p has to big enough to hold at least GetRawDataLength() bytes
{
	if (size) {
		memset(p, 0, size);
	}
	xml_memory_writer writer;
	writer.buffer = p;
	writer.remaining = size;
	m_document.save(writer);
}

bool CCryptedXmlFile::ParseData(char* data)
{
	Close();
	m_document.load_string(data);
	m_element = m_document.child(m_rootName.mb_str());
	if (!m_element) {
		Close();
	}
	return !!m_element;
}

bool CCryptedXmlFile::IsFromFutureVersion() const
{
	if (!m_element) {
		return false;
	}
	wxString const version = GetTextAttribute(m_element, "version");
	return CBuildInfo::ConvertToVersionNumber(CBuildInfo::GetVersion().c_str()) < CBuildInfo::ConvertToVersionNumber(version.c_str());
}
